"""
Utility module for JSON encoding and decoding for the agent. Handles
encoding and decoding UUID objects.
"""
import json
import uuid


class AgentEncoder(json.JSONEncoder):
    """
    Custom encoder. Encodes UUID objects as strings.
    """
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


def object_hook(json_dict):
    """
    Custom decoder hook. Searches dict for values that look like UUIDs
    and parses them.
    """
    for key, value in json_dict.items():
        # Pre-screen for potential UUIDs
        if isinstance(value, basestring) and len(value) == 36:
            try:
                # Try to parse UUID, ignore if it fails.
                json_dict[key] = uuid.UUID(value)
            except ValueError:
                pass
    return json_dict


def dumps(obj):
    """
    Custom JSON encoder that handles UUID objects.
    """
    return json.dumps(obj, cls=AgentEncoder, separators=(',', ':',),
                      sort_keys=True)


def loads(string):
    """
    Custom JSON decoder that handles UUIDs.
    """
    return json.loads(string, object_hook=object_hook)



"""
Utility functions to calculate context hashes and a stack trace.
"""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import os
import sys
import hashlib

from immunio.logger import log


THIS_DIR = os.path.dirname(__file__)

CONTEXT_CACHE = {}


def get_stack(limit=None):
    """
    Returns the current call stack.

    Based on CPython traceback.py & inspect.py but without the I/O operations
    to read the line of code.
    """
    f = sys._getframe(1)
    if limit is None:
        if hasattr(sys, 'tracebacklimit'):
            limit = sys.tracebacklimit
    stack = []
    n = 0
    while f is not None and (limit is None or n < limit):
        lineno = f.f_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        stack.append((filename, lineno, name))
        f = f.f_back
        n = n + 1
    return stack


def get_context(additional_data=None, log_context_data=False, offset=0):
    def is_immunio_frame(frame):
        if "/immunio/" in frame[0]:
            path_part = frame[0].split("/immunio/")[-1]
            path = os.path.join(THIS_DIR, path_part)
            if os.path.exists(path):
                return True

        return False

    # Change the lineno of the first frame by offset; useful in unit tests.
    stack = [frame for frame in get_stack() if not is_immunio_frame(frame)]
    stack[0] = (stack[0][0], stack[0][1] + offset, stack[0][2],)
    stack_string = "\n".join(
        ":".join(str(s) for s in frame) for frame in stack)
    cache_key = hashlib.sha1(stack_string).hexdigest()
    if cache_key in CONTEXT_CACHE:
        loose_context = CONTEXT_CACHE[cache_key]["loose_context"]
        strict_context = CONTEXT_CACHE[cache_key]["strict_context"]
        stack = CONTEXT_CACHE[cache_key]["stack"]
        loose_stack = CONTEXT_CACHE[cache_key]["loose_stack"]

        if log_context_data:
            log.info("Stack contexts from cache")
    else:
        # Use ropes as they're faster than string concatenation
        loose_stack_rope = []
        loose_context_rope = []
        stack_rope = []
        strict_context_rope = []

        # drop the top frame as it's us, but retain the rest. Immunio frames
        # are filtered by the Gem regex.
        for frame in stack[0:]:
            # Reduce paths to just use the filename part.
            strict_path = os.path.basename(frame[0])

            stack_rope.append(
                ":".join([frame[0], str(frame[1]), frame[2]]))
            strict_context_rope.append(
                ":".join([strict_path, str(frame[1]), frame[2]]))

            # Remove pathname from the loose context. The goal here
            # is to prevent upgrading package versions from changing the
            # loose context key, so for instance users don't have to
            # rebuild their whitelists every time they update a package.
            loose_context_rope.append(
                ":".join([os.path.basename(frame[0]), frame[2]]))

            # Build a second seperate rope for the stack that
            # determines ou loose context key. This includes filenames
            # for usability -- just method names not being very good
            # for display purposes...
            loose_stack_rope.append(
                ":".join([frame[0], frame[2]]))

        stack = "\n".join(stack_rope)
        strict_stack = "\n".join(strict_context_rope)
        loose_stack = "\n".join(loose_stack_rope)

        if log_context_data:
            log.info("Strict context stack:\n%s", strict_stack)
            log.info("Loose context stack:\n%s", loose_stack)

        strict_context = hashlib.sha1(strict_stack).hexdigest()
        loose_context = hashlib.sha1(
            "\n".join(loose_context_rope)).hexdigest()

        CONTEXT_CACHE[cache_key] = {
            "strict_context": strict_context,
            "loose_context": loose_context,
            "stack": stack,
            "loose_stack": loose_stack,
        }

    # Mix in additional context data
    if additional_data:
        if log_context_data:
            log.info("Additional context data:\n%s", additional_data)

        strict_context = hashlib.sha1(
            strict_context + additional_data).hexdigest()

    return strict_context, loose_context, stack, loose_stack



from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from pkg_resources import resource_filename


__agent_name__ = "agent-python"
__version__ = "1.1.0"
__vm_version__ = "2.2.0"


__ca_file__ = resource_filename(__name__, "immunio_ca.crt")


# Hold singleton copies of key objects for this app
_global_config = None
_global_plugin_manager = None
_global_agent = None


def auto_start():
    """
    Called automatically if you `import immunio.start`. Loads the agent
    config and does the MonkeyPatching. The agent will be created and
    started automatically on the first request.
    """
    config = get_config()
    if not config.agent_enabled:
        # If agent is not enabled, do nothing
        return
    # Do MonkeyPatching
    get_plugin_manager()


def get_config():
    """
    Load the Agent Config into a singleton.
    """
    from immunio.config import Config
    global _global_config

    if not _global_config:
        _global_config = Config()
    return _global_config


def get_plugin_manager():
    """
    Load PluginManager into a singleton and actually do the MonkeyPatching.
    """
    from immunio.plugin_manager import PluginManager
    global _global_plugin_manager

    if not _global_plugin_manager:
        config = get_config()
        _global_plugin_manager = PluginManager(config, get_agent_func=get_agent)
    return _global_plugin_manager


def get_agent():
    """
    Shortcut function for accessing a singleton agent.
    """
    # Import inside function to avoid circular import.
    from immunio import agent
    global _global_agent

    if not _global_agent:
        _global_agent = agent.Agent(get_config(), get_plugin_manager())
        _global_agent.start()
    return _global_agent


def start():
    """
    Manual function to create an agent and start it up. Returns
    a reference to the agent so it can be used in a call to
    `agent.wrap_wsgi_app(original_app)`.

    This is only required for frameworks where we don't automatically wrap
    the produced wsgi app. In most cases you should just
    `import immunio.start` at the top of your application entry file.
    """
    return get_agent()


def report_failed_login_attempt(username, reason=None):
    """
    Inform Immunio of a failed login attempt.
    """
    global _global_agent
    # If we don't have an enabled agent, do nothing.
    if not _global_agent or not _global_agent.enabled:
        return

    _global_agent.run_hook("immunio_api", "authenticate", {
        "is_valid": False,
        "username": username,
        "reason": reason,
    })


def report_custom_threat(threat_name, message, metadata=None):
    """
    Inform Immunio of custom threat for your app.
    """
    global _global_agent
    # If we don't have an enabled agent, do nothing.
    if not _global_agent or not _global_agent.enabled:
        return

    if not (isinstance(threat_name, str) or isinstance(threat_name, unicode)):
        raise ValueError(
            "`threat_name` must be str or unicode, not %r" % threat_name)

    if not (isinstance(message, str) or isinstance(message, unicode)):
        raise ValueError(
            "`message` must be str or unicode, not %r" % message)

    if metadata is None:
        metadata = {}

    if not isinstance(metadata, dict):
        raise ValueError("`metadata` must be a dict, not %r" % metadata)

    _global_agent.run_hook("immunio_api", "custom_threat", {
        "threat_name": threat_name,
        "message": message,
        "display_meta": metadata,
    })



from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from immunio.logger import log
from immunio.patcher import monkeypatch


# Set name so plugin can be enabled and disabled.
NAME = "flask"


def add_hooks(run_hook, get_agent_func=None):
    """
    Add our hooks into the flask library functions.
    """
    try:
        import flask
    except ImportError:
        return None

    # Install a hook to capture newly created wsgi apps and wrap them.
    hook_flask_app(run_hook, get_agent_func)

    meta = {
        "version": flask.__version__,
    }

    return meta


def hook_flask_app(run_hook, get_agent_func):
    """
    Wrap the `Flask()` __init__ function so we can wrap each WSGI
    app as it is produced. This also creates the Agent if it hasn't been
    created yet.
    """
    try:
        import flask
    except ImportError:
        return

    # If we don't have a `get_agent_func()` defined the app will be
    # wrapped elsewhere.
    if not get_agent_func:
        return

    # For new-style classes, special methods like __call__ are looked up
    # on the class directly, not the instance. This means it can't be
    # overridden normally. Instead, we have to override it on the class
    # to get it to check an instance method to allow us to override it
    # from the __init__ monkeypatch below.
    def immunio_call(self, *args, **kwargs):
        """
        Simply duplicate the behaviour of the original __call__ and proxy
        everything to the internal `wsgi_app` method. This stub will be
        wrapped by the Agent during the `__init__` monkeypatch below.
        """
        return self.wsgi_app(*args, **kwargs)
    flask.Flask._immunio_call = immunio_call

    @monkeypatch(flask.Flask, "__call__")
    def _call(orig, flask_self, *args, **kwargs):
        """
        We patch __call__ here because it is impossible to patch it on a
        per-instance basis. This mokeypatch on the class simply proxies
        through the stub `_immunio_call` defined above. Since the
        `_immunio_call` method is "normal", it can be overridden on the
        Flask instance during the __init__ monkeypatch below.
        """
        log.debug("Call to patched __call__(%(args)s, %(kwargs)s)", {
            "args": args,
            "kwargs": kwargs,
        })
        # Always call the immunio_call stub defined above. Since the stub
        # is not a special method, it can be overridden on the instance below.
        return flask_self._immunio_call(*args, **kwargs)


    @monkeypatch(flask.Flask, "__init__")
    def _flask_init(orig, flask_self, *args, **kwargs):
        """
        Here we patch the `__call__` method (via the _immunio_call stub) of
        every new Flask app. This ensures that when the app object is used
        as a WSGI callable, we already have it wrapped.
        """
        log.debug("Call to patched __init__(%(args)s, %(kwargs)s)", {
            "args": args,
            "kwargs": kwargs,
        })
        # Get the WSGI app (__init__ always returns None)
        orig(flask_self, *args, **kwargs)

        # Get or create the Immunio Agent singleton
        agent = get_agent_func()

        # Wrap the Flask app __call__ method (via _immunio_call) with Immunio.
        flask_self._immunio_call = agent.wrap_wsgi_app(flask_self._immunio_call)
